import inspect
import asyncio
from dataclasses import dataclass, fields
from typing import Type, Any, Dict, ClassVar, Callable, Awaitable, Optional
from aioquic.buffer import Buffer
from ..types import MOQTMessageType, StreamType, DatagramType
from ..utils.logger import get_logger

logger = get_logger(__name__)


@dataclass
class MOQTMessage:
    """Base class for all MOQT messages."""
    # type: Optional[int] = None - let subclass set it - annoying warnings

    def serialize(self) -> bytes:
        """Convert message to complete wire format."""
        raise NotImplementedError()

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'MOQTMessage':
        """Create message from buffer containing payload."""
        raise NotImplementedError()

    def __str__(self) -> str:
        """Generic string representation showing all fields."""
        parts = []
        class_fields = fields(self.__class__)

        for field in class_fields:
            value = getattr(self, field.name)
            # Special handling for bytes fields
            if isinstance(value, bytes):
                try:
                    str_val = value.decode('utf-8')
                except UnicodeDecodeError:
                    str_val = f"0x{value.hex()}"
            # Special handling for dicts
            elif isinstance(value, dict):
                str_val = "{" + \
                    ", ".join(f"{k}: {v}" for k, v in value.items()) + "}"
            else:
                str_val = str(value)
            parts.append(f"{field.name}={str_val}")

        return f"{self.__class__.__name__}({', '.join(parts)})"


class MOQTMessageHandler:
    """Handles parsing and routing of incoming MOQT messages."""

    # Import concrete message implementations
    from .setup import (
        ClientSetup,
        ServerSetup,
        GoAway
    )

    from .subscribe import (
        Subscribe,
        SubscribeOk,
        SubscribeError,
        SubscribeUpdate,
        Unsubscribe,
        SubscribeDone,
        MaxSubscribeId,
        SubscribesBlocked,
        TrackStatusRequest,
        TrackStatus,
    )

    from .announce import (
        Announce,
        AnnounceOk,
        AnnounceError,
        Unannounce,
        AnnounceCancel,
        SubscribeAnnounces,
        SubscribeAnnouncesOk,
        SubscribeAnnouncesError,
        UnsubscribeAnnounces
    )

    from .fetch import (
        Fetch,
        FetchOk,
        FetchError,
        FetchCancel
    )

    from .data import (
        StreamHeaderSubgroup,
        FetchHeader,
        ObjectDatagram,
        ObjectDatagramStatus
    )

    # MOQT message types to class map
    _message_types: ClassVar[Dict[int, Type[MOQTMessage]]] = {
        # Setup messages (0x40-0x41)
        MOQTMessageType.CLIENT_SETUP: ClientSetup,         # 0x40
        MOQTMessageType.SERVER_SETUP: ServerSetup,         # 0x41

        # Subscribe messages (0x02-0x05)
        MOQTMessageType.SUBSCRIBE_UPDATE: SubscribeUpdate,  # 0x02
        MOQTMessageType.SUBSCRIBE: Subscribe,             # 0x03
        MOQTMessageType.SUBSCRIBE_OK: SubscribeOk,        # 0x04
        MOQTMessageType.SUBSCRIBE_ERROR: SubscribeError,  # 0x05

        # Announce messages (0x06-0x09)
        MOQTMessageType.ANNOUNCE: Announce,               # 0x06
        MOQTMessageType.ANNOUNCE_OK: AnnounceOk,         # 0x07
        MOQTMessageType.ANNOUNCE_ERROR: AnnounceError,   # 0x08
        MOQTMessageType.UNANNOUNCE: Unannounce,         # 0x09

        # Additional subscription messages (0x0A-0x0B)
        MOQTMessageType.UNSUBSCRIBE: Unsubscribe,        # 0x0A
        MOQTMessageType.SUBSCRIBE_DONE: SubscribeDone,   # 0x0B

        # Announce control messages (0x0C)
        MOQTMessageType.ANNOUNCE_CANCEL: AnnounceCancel,  # 0x0C

        # Status messages (0x0D-0x0E)
        MOQTMessageType.TRACK_STATUS_REQUEST: TrackStatusRequest,  # 0x0D
        MOQTMessageType.TRACK_STATUS: TrackStatus,       # 0x0E

        # Session control messages (0x10)
        MOQTMessageType.GOAWAY: GoAway,                  # 0x10

        # Subscription announce messages (0x11-0x14)
        MOQTMessageType.SUBSCRIBE_ANNOUNCES: SubscribeAnnounces,         # 0x11
        MOQTMessageType.SUBSCRIBE_ANNOUNCES_OK: SubscribeAnnouncesOk,    # 0x12
        MOQTMessageType.SUBSCRIBE_ANNOUNCES_ERROR: SubscribeAnnouncesError,  # 0x13
        MOQTMessageType.UNSUBSCRIBE_ANNOUNCES: UnsubscribeAnnounces,    # 0x14

        # Subscribe control messages (0x15, 0x1A)
        MOQTMessageType.MAX_SUBSCRIBE_ID: MaxSubscribeId,      # 0x15
        MOQTMessageType.SUBSCRIBES_BLOCKED: SubscribesBlocked,  # 0x1A

        # Fetch messages (0x16-0x19)
        MOQTMessageType.FETCH: Fetch,                    # 0x16
        MOQTMessageType.FETCH_CANCEL: FetchCancel,       # 0x17
        MOQTMessageType.FETCH_OK: FetchOk,               # 0x18
        MOQTMessageType.FETCH_ERROR: FetchError,         # 0x19
    }

    # Mapping of stream types to message classes
    _stream_types: ClassVar[Dict[int, Type[MOQTMessage]]] = {
        StreamType.STREAM_HEADER_SUBGROUP: StreamHeaderSubgroup,
        StreamType.FETCH_HEADER: FetchHeader,
    }

    # Mapping of datagram types to message classes
    _datagram_types: ClassVar[Dict[int, Type[MOQTMessage]]] = {
        DatagramType.OBJECT_DATAGRAM: ObjectDatagram,
        DatagramType.OBJECT_DATAGRAM_STATUS: ObjectDatagramStatus,
    }

    def __init__(self, protocol: Any):
        self.protocol = protocol
        self._custom_handlers: Dict[int, Callable[[
            MOQTMessage], Awaitable[None]]] = {}
        self._tasks = set()

    def handle_control_message(self, data: bytes) -> Optional[MOQTMessage]:
        """Process an incoming message."""
        if not data:
            logger.warning(
                f"{inspect.currentframe().f_code.co_name}:  empty message data")
            return None

        try:
            buffer = Buffer(data=data)
            msg_type = buffer.pull_uint_var()
            length = buffer.pull_uint_var()

            # Look up message class
            message_class = self._message_types.get(msg_type)
            if message_class is None:
                raise ValueError(
                    f"MOQT event: unknown control message type: {hex(msg_type)}")

            # Deserialize message
            logger.debug(
                f"MOQT event: control message: {message_class} {message_class.__class__} ({hex(msg_type)})")
            message = message_class.deserialize(buffer)
            logger.debug(f"MOQT event: control message: {message}")

            # Schedule handler if one exists
            handler = self._custom_handlers.get(msg_type)
            if handler:
                task = asyncio.create_task(handler(message))
                self._tasks.add(task)

            return message

        except Exception as e:
            logger.error(
                f"{inspect.currentframe().f_code.co_name}: error handling control message: {e}")
            raise

    def handle_data_message(self, stream_id: int, data: bytes) -> None:
        """Process incoming data messages (not control messages)."""
        if not data:
            logger.error(
                f"MOQT event: stream {stream_id}: message contains no data")
            return

        try:
            # Handle STREAM_HEADER messages
            if len(data) > 0:
                stream_type = data[0]
                logger.info(
                    f"MOQT: stream {stream_id}: type: {hex(stream_type)}")

                if stream_type == StreamType.STREAM_HEADER_SUBGROUP:
                    self._handle_subgroup_header(data)
                    return

            # Handle object data
            self._handle_object_data(data)

        except Exception as e:
            logger.error(
                f"{inspect.currentframe().f_code.co_name}: error handling data message: {e}")

    def _handle_subgroup_header(self, data: bytes) -> None:
        """Process subgroup header messages."""
        if len(data) >= 13:
            group_id = int.from_bytes(data[1:5], 'big')
            subgroup_id = int.from_bytes(data[5:9], 'big')
            priority = data[9]
            logger.info("  Message type: STREAM_HEADER_SUBGROUP")
            logger.info(f"  Group: {group_id}")
            logger.info(f"  Subgroup: {subgroup_id}")
            logger.info(f"  Priority: {priority}")

    def _handle_object_data(self, data: bytes) -> None:
        """Process object data messages."""
        group_id = int.from_bytes(data[0:4], 'big')
        subgroup_id = int.from_bytes(data[4:8], 'big')
        object_id = int.from_bytes(data[8:12], 'big')
        payload = data[12:]

        # Update statistics
        self._groups[group_id]['objects'] += 1
        self._groups[group_id]['subgroups'].add(subgroup_id)

        logger.info("  Object received:")
        logger.info(f"    Group: {group_id}")
        logger.info(f"    Subgroup: {subgroup_id}")
        logger.info(f"    Object: {object_id}")
        logger.info(f"    Payload size: {len(payload)}")

    def register_handler(self, msg_type: int,
                         handler: Callable[[MOQTMessage], Awaitable[None]]) -> None:
        """Register an async message handler for a specific message type."""
        self._custom_handlers[msg_type] = handler



from dataclasses import dataclass
from enum import IntEnum
from typing import Optional, Dict, ClassVar, Type, Union
from aioquic.buffer import Buffer
from .base import MOQTMessage
from ..types import StreamType, DatagramType, ObjectStatus, ForwardingPreference

@dataclass
class StreamHeaderSubgroup(MOQTMessage):
    """Stream header for subgroup data."""
    track_alias: int
    group_id: int
    subgroup_id: int
    publisher_priority: int

    def __post_init__(self):
        self.type = StreamType.STREAM_HEADER_SUBGROUP

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)
        payload = Buffer(capacity=32)

        payload.push_uint_var(self.track_alias)
        payload.push_uint_var(self.group_id)
        payload.push_uint_var(self.subgroup_id)
        payload.push_uint8(self.publisher_priority)

        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'StreamHeaderSubgroup':
        track_alias = buffer.pull_uint_var()
        group_id = buffer.pull_uint_var()
        subgroup_id = buffer.pull_uint_var()
        publisher_priority = buffer.pull_uint8()
        
        return cls(
            track_alias=track_alias,
            group_id=group_id,
            subgroup_id=subgroup_id,
            publisher_priority=publisher_priority
        )

@dataclass
class FetchHeader(MOQTMessage):
    """Stream header for fetch data."""
    subscribe_id: int

    def __post_init__(self):
        self.type = StreamType.FETCH_HEADER

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)
        payload = Buffer(capacity=32)

        payload.push_uint_var(self.subscribe_id)

        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'FetchHeader':
        subscribe_id = buffer.pull_uint_var()
        return cls(subscribe_id=subscribe_id)

@dataclass
class ObjectDatagram(MOQTMessage):
    """Object datagram message."""
    track_alias: int
    group_id: int
    object_id: int
    publisher_priority: int
    payload: bytes = b''

    def __post_init__(self):
        self.type = DatagramType.OBJECT_DATAGRAM

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32 + len(self.payload))

        buf.push_uint_var(self.track_alias)
        buf.push_uint_var(self.group_id)
        buf.push_uint_var(self.object_id)
        buf.push_uint8(self.publisher_priority)
        buf.push_bytes(self.payload)

        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'ObjectDatagram':
        track_alias = buffer.pull_uint_var()
        group_id = buffer.pull_uint_var()
        object_id = buffer.pull_uint_var()
        publisher_priority = buffer.pull_uint8()
        payload = buffer.pull_bytes(buffer.capacity - buffer.tell())

        return cls(
            track_alias=track_alias,
            group_id=group_id,
            object_id=object_id,
            publisher_priority=publisher_priority,
            payload=payload
        )

@dataclass
class ObjectDatagramStatus(MOQTMessage):
    """Object datagram status message."""
    track_alias: int
    group_id: int
    object_id: int
    publisher_priority: int
    status: ObjectStatus

    def __post_init__(self):
        self.type = DatagramType.OBJECT_DATAGRAM_STATUS

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)

        buf.push_uint_var(self.track_alias)
        buf.push_uint_var(self.group_id)
        buf.push_uint_var(self.object_id)
        buf.push_uint8(self.publisher_priority)
        buf.push_uint_var(self.status)

        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'ObjectDatagramStatus':
        track_alias = buffer.pull_uint_var()
        group_id = buffer.pull_uint_var()
        object_id = buffer.pull_uint_var()
        publisher_priority = buffer.pull_uint8()
        status = ObjectStatus(buffer.pull_uint_var())

        return cls(
            track_alias=track_alias,
            group_id=group_id,
            object_id=object_id,
            publisher_priority=publisher_priority,
            status=status
        )

@dataclass
class StreamObject(MOQTMessage):
    """Object within a subgroup stream."""
    object_id: int
    payload_length: int
    status: Optional[ObjectStatus] = None
    payload: bytes = b''

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32 + len(self.payload))

        buf.push_uint_var(self.object_id)
        buf.push_uint_var(self.payload_length)

        if self.payload_length == 0 and self.status is not None:
            buf.push_uint_var(self.status)
        elif self.payload:
            buf.push_bytes(self.payload)

        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'StreamObject':
        object_id = buffer.pull_uint_var()
        payload_length = buffer.pull_uint_var()
        
        status = None
        payload = b''
        
        if payload_length == 0:
            status = ObjectStatus(buffer.pull_uint_var())
        else:
            payload = buffer.pull_bytes(payload_length)

        return cls(
            object_id=object_id,
            payload_length=payload_length,
            status=status,
            payload=payload
        )

@dataclass
class FetchObject(MOQTMessage):
    """Object within a fetch stream."""
    group_id: int
    subgroup_id: int
    object_id: int
    publisher_priority: int
    payload_length: int
    status: Optional[ObjectStatus] = None
    payload: bytes = b''

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32 + len(self.payload))

        buf.push_uint_var(self.group_id)
        buf.push_uint_var(self.subgroup_id)
        buf.push_uint_var(self.object_id)
        buf.push_uint8(self.publisher_priority)
        buf.push_uint_var(self.payload_length)

        if self.payload_length == 0 and self.status is not None:
            buf.push_uint_var(self.status)
        elif self.payload:
            buf.push_bytes(self.payload)

        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'FetchObject':
        group_id = buffer.pull_uint_var()
        subgroup_id = buffer.pull_uint_var()
        object_id = buffer.pull_uint_var()
        publisher_priority = buffer.pull_uint8()
        payload_length = buffer.pull_uint_var()

        status = None
        payload = b''

        if payload_length == 0:
            status = ObjectStatus(buffer.pull_uint_var())
        else:
            payload = buffer.pull_bytes(payload_length)

        return cls(
            group_id=group_id,
            subgroup_id=subgroup_id,
            object_id=object_id,
            publisher_priority=publisher_priority,
            payload_length=payload_length,
            status=status,
            payload=payload
        )




#!/usr/bin/env python3

import argparse
import logging
import ssl

import asyncio
from aioquic.h3.connection import H3_ALPN
from aioquic.quic.configuration import QuicConfiguration
from aiomoqt.types import ParamType
from aiomoqt.client import MOQTClient, connect
from aiomoqt.utils.logger import get_logger, set_log_level, QuicDebugLogger


def parse_args():
    parser = argparse.ArgumentParser(description='MOQT WebTransport Client')
    parser.add_argument('--host', type=str, default='localhost',
                        help='Host to connect to')
    parser.add_argument('--port', type=int, default=4433,
                        help='Port to connect to')
    parser.add_argument('--namespace', type=str, required=True,
                        help='Track namespace')
    parser.add_argument('--trackname', type=str, required=True,
                        help='Track name')
    parser.add_argument('--endpoint', type=str, required=True,
                        help='MOQT WT endpoint')
    parser.add_argument('--timeout', type=int, default=30,
                        help='How long to run before unsubscribing (seconds)')
    parser.add_argument('--debug', action='store_true',
                        help='Enable debug output')
    return parser.parse_args()


async def main(host: str, port: int, endpoint: str, namespace: str, trackname: str, timeout: int,  debug: bool):
    log_level = logging.DEBUG if debug else logging.INFO
    set_log_level(log_level)
    logger = get_logger(__name__)
    try:
        configuration = QuicConfiguration(
            alpn_protocols=H3_ALPN,
            is_client=True,
            verify_mode=ssl.CERT_NONE,
            quic_logger=QuicDebugLogger() if debug else None,
            secrets_log_file=open("/tmp/keylog.client.txt",
                                  "a") if debug else None
        )

        client = MOQTClient(host, port, endpoint=endpoint,
                            configuration=configuration, debug=debug)

        async with client.connect() as client_session:
            await asyncio.wait_for(client_session.initialize(), timeout=30)
            client_session.subscribe_announces(namespace_prefix=namespace,
                                    parameters={ParamType.AUTHORIZATION_INFO: b"auth-token-123"})
            await asyncio.sleep(timeout)
            client_session.subscribe(
                namespace=namespace, track_name=trackname)
            await asyncio.sleep(timeout)
            logger.info(f"still in context: {client_session}")

        logger.info(f"finished: client: {client}")

    except asyncio.TimeoutError:
        logger.error("Operation timed out")
        raise

if __name__ == "__main__":
    args = parse_args()
    asyncio.run(main(
        host=args.host,
        port=args.port,
        endpoint=args.endpoint,
        namespace=args.namespace,
        trackname=args.trackname,
        timeout=args.timeout,
        debug=args.debug
    ))



import asyncio
import ssl
import os
from typing import Optional, Dict, Any
from dataclasses import dataclass
from pathlib import Path

from aioquic.asyncio.server import serve
from aioquic.h3.connection import H3_ALPN
from aioquic.quic.configuration import QuicConfiguration
from aioquic.tls import load_private_key, load_certificate_chain

from .protocol import MOQTProtocol
from .types import SessionCloseCode
from .utils.logger import get_logger

logger = get_logger(__name__)


@dataclass
class ServerConfig:
    """MOQT Server configuration."""
    host: str = "localhost"
    port: int = 4433
    cert_path: str = "cert.pem"
    key_path: str = "key.pem"
    verify_mode: int = ssl.CERT_NONE
    max_datagram_size: Optional[int] = None


class MOQTServer(MOQTProtocol):
    """MOQT server implementation."""

    async def initialize(self, **kwargs) -> None:
        """Initialize server-side of MOQT session."""
        # WebTransport/H3 is already initialized by this point
        # Wait for CLIENT_SETUP
        try:
            await asyncio.wait_for(self._wt_session.wait(), timeout=30.0)
            logger.info("WebTransport session established")
        except asyncio.TimeoutError:
            logger.error("WebTransport session establishment timeout")
            raise

        # Send SERVER_SETUP
        logger.info("Sending SERVER_SETUP")
        await self.send_message(
            self.message_builder.server_setup(version=0xff000007)
        )

        self._moqt_session.set()
        logger.info("MOQT session setup complete")


def load_server_config(config: ServerConfig) -> QuicConfiguration:
    """Load QUIC server configuration with certificates."""
    if not os.path.exists(config.cert_path) or not os.path.exists(config.key_path):
        logger.error(f"Certificate ({config.cert_path}) or key ({
                     config.key_path}) not found")
        raise FileNotFoundError("Certificate or key file not found")

    quic_config = QuicConfiguration(
        alpn_protocols=H3_ALPN,
        is_client=False,
        max_datagram_size=config.max_datagram_size,
    )

    try:
        quic_config.load_cert_chain(config.cert_path, config.key_path)
    except Exception as e:
        logger.error(f"Failed to load certificates: {e}")
        raise

    return quic_config


async def create_server(config: ServerConfig) -> asyncio.AbstractServer:
    """Create and start a MOQT server."""
    quic_config = load_server_config(config)

    return await serve(
        host=config.host,
        port=config.port,
        configuration=quic_config,
        create_protocol=MOQTServer,
    )

# Example usage:
"""
# Generate test certificates (for development only):
from .utils.certs import generate_test_certs

# Server setup and run:
async def run_server():
    config = ServerConfig(
        host="localhost",
        port=4433,
        cert_path="cert.pem",
        key_path="key.pem"
    )
    
    # Generate test certs if needed
    if not os.path.exists(config.cert_path):
        generate_test_certs(config.cert_path, config.key_path)
    
    server = await create_server(config)
    
    try:
        await asyncio.Future()  # run forever
    finally:
        server.close()
        await server.wait_closed()

if __name__ == "__main__":
    asyncio.run(run_server())
"""



from dataclasses import dataclass
from typing import Dict, Optional, List, Tuple
from aioquic.buffer import Buffer
from .base import MOQTMessage 
from ..types import MOQTMessageType
from ..utils.logger import get_logger

logger = get_logger(__name__)

@dataclass
class Announce(MOQTMessage):
    """ANNOUNCE message for advertising a track namespace."""
    namespace: Tuple[bytes, ...] = None  # Track namespace as a tuple of bytes
    parameters: Dict[int, bytes] = None  # Optional parameters

    def __post_init__(self):
        self.type = MOQTMessageType.ANNOUNCE

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)
        payload = Buffer(capacity=32)

        # Serialize namespace
        payload.push_uint_var(len(self.namespace))
        for part in self.namespace:
            payload.push_uint_var(len(part))
            payload.push_bytes(part)

        # Serialize parameters
        payload.push_uint_var(len(self.parameters))
        for param_id, param_value in self.parameters.items():
            payload.push_uint_var(param_id)
            payload.push_uint_var(len(param_value))
            payload.push_bytes(param_value)

        # Build final message
        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'Announce':
        # Deserialize namespace tuple
        tuple_len = buffer.pull_uint_var()
        namespace = []
        for _ in range(tuple_len):
            part_len = buffer.pull_uint_var()
            namespace.append(buffer.pull_bytes(part_len))
        namespace = tuple(namespace)  # Convert to tuple
        
        # Deserialize parameters
        params = {}
        param_count = buffer.pull_uint_var()
        for _ in range(param_count):
            param_id = buffer.pull_uint_var()
            param_len = buffer.pull_uint_var()
            param_value = buffer.pull_bytes(param_len)
            params[param_id] = param_value
            
        return cls(namespace=namespace, parameters=params)

@dataclass
class AnnounceOk(MOQTMessage):
    """ANNOUNCE_OK response message."""
    namespace: Tuple[bytes, ...]

    def __post_init__(self):
        self.type = MOQTMessageType.ANNOUNCE_OK

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)
        payload = Buffer(capacity=32)

        payload.push_uint_var(len(self.namespace))
        for part in self.namespace:
            payload.push_uint_var(len(part))
            payload.push_bytes(part)

        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'AnnounceOk':
        tuple_len = buffer.pull_uint_var()
        namespace = []
        for _ in range(tuple_len):
            part_len = buffer.pull_uint_var()
            namespace.append(buffer.pull_bytes(part_len))
        return cls(namespace=tuple(namespace))

@dataclass
class AnnounceError(MOQTMessage):
    """ANNOUNCE_ERROR response message."""
    namespace: Tuple[bytes, ...]
    error_code: int
    reason: str

    def __post_init__(self):
        self.type = MOQTMessageType.ANNOUNCE_ERROR

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)
        payload = Buffer(capacity=32)

        payload.push_uint_var(len(self.namespace))
        for part in self.namespace:
            payload.push_uint_var(len(part))
            payload.push_bytes(part)
            
        payload.push_uint_var(self.error_code)
        
        reason_bytes = self.reason.encode()
        payload.push_uint_var(len(reason_bytes))
        payload.push_bytes(reason_bytes)

        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'AnnounceError':
        tuple_len = buffer.pull_uint_var()
        namespace = []
        for _ in range(tuple_len):
            part_len = buffer.pull_uint_var()
            namespace.append(buffer.pull_bytes(part_len))
        
        error_code = buffer.pull_uint_var()
        reason_len = buffer.pull_uint_var()
        reason = buffer.pull_bytes(reason_len).decode()
        return cls(namespace=tuple(namespace), error_code=error_code, reason=reason)

@dataclass
class Unannounce(MOQTMessage):
    """UNANNOUNCE message to withdraw track namespace."""
    namespace: Tuple[bytes, ...]

    def __post_init__(self):
        self.type = MOQTMessageType.UNANNOUNCE

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)
        payload = Buffer(capacity=32)

        payload.push_uint_var(len(self.namespace))
        for part in self.namespace:
            payload.push_uint_var(len(part))
            payload.push_bytes(part)

        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'Unannounce':
        tuple_len = buffer.pull_uint_var()
        namespace = []
        for _ in range(tuple_len):
            part_len = buffer.pull_uint_var()
            namespace.append(buffer.pull_bytes(part_len))
        return cls(namespace=tuple(namespace))

@dataclass
class AnnounceCancel(MOQTMessage):
    """ANNOUNCE_CANCEL message to withdraw announcement acceptance."""
    namespace: Tuple[bytes, ...]
    error_code: int
    reason: str

    def __post_init__(self):
        self.type = MOQTMessageType.ANNOUNCE_CANCEL

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)
        payload = Buffer(capacity=32)

        payload.push_uint_var(len(self.namespace))
        for part in self.namespace:
            payload.push_uint_var(len(part))
            payload.push_bytes(part)
            
        payload.push_uint_var(self.error_code)
        
        reason_bytes = self.reason.encode()
        payload.push_uint_var(len(reason_bytes))
        payload.push_bytes(reason_bytes)

        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'AnnounceCancel':
        tuple_len = buffer.pull_uint_var()
        namespace = []
        for _ in range(tuple_len):
            part_len = buffer.pull_uint_var()
            namespace.append(buffer.pull_bytes(part_len))
        
        error_code = buffer.pull_uint_var()
        reason_len = buffer.pull_uint_var()
        reason = buffer.pull_bytes(reason_len).decode()
        return cls(namespace=tuple(namespace), error_code=error_code, reason=reason)

@dataclass
class SubscribeAnnounces(MOQTMessage):
    """SUBSCRIBE_ANNOUNCES message to subscribe to announcements."""
    namespace_prefix: Tuple[bytes, ...]  # Track namespace prefix as tuple
    parameters: Dict[int, bytes]

    def __post_init__(self):
        self.type = MOQTMessageType.SUBSCRIBE_ANNOUNCES

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)
        payload = Buffer(capacity=32)

        payload.push_uint_var(len(self.namespace_prefix))
        for part in self.namespace_prefix:
            payload.push_uint_var(len(part))
            payload.push_bytes(part)

        payload.push_uint_var(len(self.parameters))
        for param_id, param_value in self.parameters.items():
            payload.push_uint_var(param_id)
            payload.push_uint_var(len(param_value))
            payload.push_bytes(param_value)

        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'SubscribeAnnounces':
        tuple_len = buffer.pull_uint_var()
        namespace_prefix = []
        for _ in range(tuple_len):
            part_len = buffer.pull_uint_var()
            namespace_prefix.append(buffer.pull_bytes(part_len))

        params = {}
        param_count = buffer.pull_uint_var()
        for _ in range(param_count):
            param_id = buffer.pull_uint_var()
            param_len = buffer.pull_uint_var()
            param_value = buffer.pull_bytes(param_len)
            params[param_id] = param_value

        return cls(namespace_prefix=tuple(namespace_prefix), parameters=params)

@dataclass
class SubscribeAnnouncesOk(MOQTMessage):
    """SUBSCRIBE_ANNOUNCES_OK response message."""
    namespace_prefix: Tuple[bytes, ...]

    def __post_init__(self):
        self.type = MOQTMessageType.SUBSCRIBE_ANNOUNCES_OK

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)
        payload = Buffer(capacity=32)

        payload.push_uint_var(len(self.namespace_prefix))
        for part in self.namespace_prefix:
            payload.push_uint_var(len(part))
            payload.push_bytes(part)

        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'SubscribeAnnouncesOk':
        tuple_len = buffer.pull_uint_var()
        namespace_prefix = []
        for _ in range(tuple_len):
            part_len = buffer.pull_uint_var()
            namespace_prefix.append(buffer.pull_bytes(part_len))
        return cls(namespace_prefix=tuple(namespace_prefix))

@dataclass
class SubscribeAnnouncesError(MOQTMessage):
    """SUBSCRIBE_ANNOUNCES_ERROR response message."""
    namespace_prefix: Tuple[bytes, ...]
    error_code: int
    reason: str

    def __post_init__(self):
        self.type = MOQTMessageType.SUBSCRIBE_ANNOUNCES_ERROR

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)
        payload = Buffer(capacity=32)

        payload.push_uint_var(len(self.namespace_prefix))
        for part in self.namespace_prefix:
            payload.push_uint_var(len(part))
            payload.push_bytes(part)
            
        payload.push_uint_var(self.error_code)
        
        reason_bytes = self.reason.encode()
        payload.push_uint_var(len(reason_bytes))
        payload.push_bytes(reason_bytes)

        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'SubscribeAnnouncesError':
        tuple_len = buffer.pull_uint_var()
        namespace_prefix = []
        for _ in range(tuple_len):
            part_len = buffer.pull_uint_var()
            namespace_prefix.append(buffer.pull_bytes(part_len))
        
        error_code = buffer.pull_uint_var()
        reason_len = buffer.pull_uint_var()
        reason = buffer.pull_bytes(reason_len).decode()
        return cls(namespace_prefix=tuple(namespace_prefix), error_code=error_code, reason=reason)

@dataclass 
class UnsubscribeAnnounces(MOQTMessage):
    """UNSUBSCRIBE_ANNOUNCES message."""
    namespace_prefix: Tuple[bytes, ...]

    def __post_init__(self):
        self.type = MOQTMessageType.UNSUBSCRIBE_ANNOUNCES

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)
        payload = Buffer(capacity=32)

        payload.push_uint_var(len(self.namespace_prefix))
        for part in self.namespace_prefix:
            payload.push_uint_var(len(part))
            payload.push_bytes(part)

        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'UnsubscribeAnnounces':
        tuple_len = buffer.pull_uint_var()
        namespace_prefix = []
        for _ in range(tuple_len):
            part_len = buffer.pull_uint_var()
            namespace_prefix.append(buffer.pull_bytes(part_len))
        return cls(namespace_prefix=tuple(namespace_prefix))


from dataclasses import dataclass
from typing import Dict, List, Optional
from aioquic.buffer import Buffer
from .base import MOQTMessage
from ..types import MOQTMessageType, SetupParamType
from ..utils.logger import get_logger

logger = get_logger(__name__)


@dataclass
class ServerSetup(MOQTMessage):
    """SERVER_SETUP message for accepting MOQT session."""
    selected_version: int = None
    parameters: Dict[int, bytes] = None

    def __post_init__(self):
        self.type = MOQTMessageType.SERVER_SETUP

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)
        payload = Buffer(capacity=32)

        # Add selected version
        payload.push_uint_var(self.selected_version)

        # Add parameters
        payload.push_uint_var(len(self.parameters))
        for param_id, param_value in self.parameters.items():
            payload.push_uint_var(param_id)
            payload.push_uint_var(len(param_value))
            payload.push_bytes(param_value)

        # Build final message
        buf.push_uint_var(self.type)  # SERVER_SETUP type
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'ServerSetup':
        """Handle SERVER_SETUP message."""
        version = buffer.pull_uint_var()
        param_count = buffer.pull_uint_var()

        logger.info(
            f"SERVER_SETUP: version: {hex(version)} params: {param_count} ")
        params = {}
        for _ in range(param_count):
            param_id = buffer.pull_uint_var()
            param_len = buffer.pull_uint_var()
            param_value = buffer.pull_bytes(param_len)
            if (param_id == SetupParamType.MAX_SUBSCRIBER_ID):
                id = "MAX_SUBSCRIBER_ID"
                param_value = Buffer(data=param_value).pull_uint_var()
            elif (param_id == SetupParamType.CLIENT_ROLE):
                id = "CLIENT_ROLE"
            elif (param_id == SetupParamType.ENDPOINT_PATH):
                id = "ENDPOINT_PATH"
            else:
                id = "UNKNOWN"
                logger.error(
                    f"_handle_server_setup: received unknown setup param type: {hex(param_id)}")
            logger.debug(
                f"  param: id: {id} ({hex(param_id)}) len: {param_len} val: {param_value}")
            params[param_id] = param_value
        return cls(selected_version=version, parameters=params)
        # self.protocol._moqt_session.set()


@dataclass
class ClientSetup(MOQTMessage):
    """CLIENT_SETUP message for initializing MOQT session."""
    versions: List[int] = None
    parameters: Dict[int, bytes] = None

    def __post_init__(self):
        self.type = MOQTMessageType.CLIENT_SETUP

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)
        payload = Buffer(capacity=32)

        # Add versions
        payload.push_uint_var(len(self.versions))
        for version in self.versions:
            payload.push_uint_var(version)

        # Add parameters
        payload.push_uint_var(len(self.parameters))
        for param_id, param_value in self.parameters.items():
            payload.push_uint_var(param_id)
            payload.push_uint_var(len(param_value))
            payload.push_bytes(param_value)

        # Build final message
        buf.push_uint_var(self.type)  # CLIENT_SETUP type
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> None:
        """Handle CLIENT_SETUP message."""

        versions = []
        version_count = buffer.pull_uint_var()
        for _ in range(version_count):
            versions.append(buffer.pull_uint_var())

        param_count = buffer.pull_uint_var()

        logger.info(
            f"CLIENT_SETUP: version: {versions} params: {param_count} ")
        params = {}
        for _ in range(param_count):
            param_id = buffer.pull_uint_var()
            param_len = buffer.pull_uint_var()
            param_value = buffer.pull_bytes(param_len)
            if (param_id == SetupParamType.MAX_SUBSCRIBER_ID):
                id = "MAX_SUBSCRIBER_ID"
                param_value = Buffer(data=param_value).pull_uint_var()
            elif (param_id == SetupParamType.CLIENT_ROLE):
                id = "CLIENT_ROLE"
            elif (param_id == SetupParamType.ENDPOINT_PATH):
                id = "ENDPOINT_PATH"
            else:
                id = "UNKNOWN"
                logger.error(
                    f"_handle_server_setup: received unknown setup param type: {hex(param_id)}")
            params[param_id] = param_value
            logger.debug(
                f"  param: id: {id} ({hex(param_id)}) len: {param_len} val: {param_value}")

            return cls(versions=versions, parameters=params)
        

@dataclass
class GoAway(MOQTMessage):
    new_session_uri: str = None

    def __post_init__(self):
        self.type = MOQTMessageType.GOAWAY

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32 + len(self.new_session_uri))

        # Calculate payload size
        payload_size = 1  # uri length varint
        payload_size += len(self.new_session_uri.encode())  # uri bytes

        # Write message
        buf.push_uint_var(self.type)
        buf.push_uint_var(payload_size)

        # Write URI length and data
        uri_bytes = self.new_session_uri.encode()
        buf.push_uint_var(len(uri_bytes))
        buf.push_bytes(uri_bytes)

        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'GoAway':
        uri_len = buffer.pull_uint_var()
        uri = buffer.pull_bytes(uri_len).decode()
        return cls(new_session_uri=uri)



from ..types import MOQTMessageType, TrackStatusCode
from typing import Tuple, Dict, Optional
from dataclasses import dataclass
from typing import Dict, List, Optional
from aioquic.buffer import Buffer
from .base import MOQTMessage

from ..utils.logger import get_logger

logger = get_logger(__name__)


# def subscribe(self, subscribe_id: int, track_alias: int, namespace: bytes,
#               track_name: bytes, priority: int = 128,
#               direction: int = 0x1, filter_type: int = 0x1,
#               start_group: Optional[int] = None,
#               start_object: Optional[int] = None,
#               end_group: Optional[int] = None,
#               parameters: Optional[Dict[int, bytes]] = None) -> bytes:
#     """Create a SUBSCRIBE message."""
#     msg = Subscribe(
#         type=MOQTMessageType.SUBSCRIBE,
#         subscribe_id=subscribe_id,
#         track_alias=track_alias,
#         namespace=namespace,
#         track_name=track_name,
#         priority=priority,
#         direction=direction,
#         filter_type=filter_type,
#         start_group=start_group,
#         start_object=start_object,
#         end_group=end_group,
#         parameters=parameters
#     )
#     return msg.serialize()


# def unsubscribe(self, subscribe_id: int) -> bytes:
#     """Create an UNSUBSCRIBE message."""
#     msg = Unsubscribe(
#         type=MOQTMessageType.UNSUBSCRIBE,
#         subscribe_id=subscribe_id
#     )
#     return msg.serialize()


def _handle_subscribe_ok(self, buffer: Buffer) -> None:
    """Handle SUBSCRIBE_OK message."""
    subscribe_id = buffer.pull_uint_var()
    expires = buffer.pull_uint_var()
    group_order = buffer.pull_uint8()
    content_exists = buffer.pull_uint8()

    if content_exists == 1:
        largest_group = buffer.pull_uint_var()
        largest_object = buffer.pull_uint_var()
        logger.info(
            f"Subscription {subscribe_id} OK - Latest Group: {largest_group}, Object: {largest_object}")
    else:
        logger.info(f"Subscription {subscribe_id} OK - No content yet")


def _handle_subscribe_error(self, buffer: Buffer) -> None:
    """Handle SUBSCRIBE_ERROR message."""
    subscribe_id = buffer.pull_uint_var()
    error_code = buffer.pull_uint_var()
    reason_len = buffer.pull_uint_var()
    reason = buffer.pull_bytes(reason_len).decode('utf-8')

    logger.error(
        f"Subscription {subscribe_id} failed - Error {error_code}: {reason}")


def _handle_subscribe_done(self, buffer: Buffer) -> None:
    """Handle SUBSCRIBE_DONE message."""
    subscribe_id = buffer.pull_uint_var()
    status_code = buffer.pull_uint_var()
    stream_count = buffer.pull_uint_var()
    reason_len = buffer.pull_uint_var()
    reason = buffer.pull_bytes(reason_len).decode('utf-8')

    logger.info(
        f"Subscription {subscribe_id} done - Status {status_code}: {reason}")


@dataclass
class TrackStatusRequest(MOQTMessage):
    namespace: Tuple[bytes, ...] = None  # Tuple encoded
    track_name: bytes = None

    def __post_init__(self):
        self.type = MOQTMessageType.TRACK_STATUS_REQUEST

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)

        # Calculate payload size
        payload_size = 0
        namespace_parts = self.namespace.split(b'/')
        payload_size += 1  # num parts varint
        for part in namespace_parts:
            payload_size += 1  # part length varint
            payload_size += len(part)  # part bytes
        payload_size += 1  # track name length varint
        payload_size += len(self.track_name)  # track name bytes

        # Write message
        buf.push_uint_var(self.type)
        buf.push_uint_var(payload_size)

        # Write namespace as tuple
        buf.push_uint_var(len(namespace_parts))
        for part in namespace_parts:
            buf.push_uint_var(len(part))
            buf.push_bytes(part)

        # Write track name
        buf.push_uint_var(len(self.track_name))
        buf.push_bytes(self.track_name)

        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'TrackStatusRequest':
        # Parse namespace tuple
        namespace_parts = []
        num_parts = buffer.pull_uint_var()
        for _ in range(num_parts):
            part_len = buffer.pull_uint_var()
            part = buffer.pull_bytes(part_len)
            namespace_parts.append(part)
        namespace = b'/'.join(namespace_parts)

        # Parse track name
        track_name_len = buffer.pull_uint_var()
        track_name = buffer.pull_bytes(track_name_len)

        return cls(namespace=namespace, track_name=track_name)


@dataclass
class TrackStatus(MOQTMessage):
    namespace: Tuple[bytes, ...]  # Tuple encoded
    track_name: bytes
    status_code: TrackStatusCode
    last_group_id: int
    last_object_id: int

    def __post_init__(self):
        self.type = MOQTMessageType.TRACK_STATUS

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)

        # Calculate payload size
        payload_size = 0
        namespace_parts = self.namespace.split(b'/')
        payload_size += 1  # num parts varint
        for part in namespace_parts:
            payload_size += 1  # part length varint
            payload_size += len(part)  # part bytes
        payload_size += 1  # track name length varint
        payload_size += len(self.track_name)  # track name bytes
        payload_size += 1  # status code varint
        payload_size += 1  # last group id varint
        payload_size += 1  # last object id varint

        # Write message
        buf.push_uint_var(self.type)
        buf.push_uint_var(payload_size)

        # Write namespace as tuple
        buf.push_uint_var(len(namespace_parts))
        for part in namespace_parts:
            buf.push_uint_var(len(part))
            buf.push_bytes(part)

        # Write track name
        buf.push_uint_var(len(self.track_name))
        buf.push_bytes(self.track_name)

        # Write status info
        buf.push_uint_var(self.status_code)
        buf.push_uint_var(self.last_group_id)
        buf.push_uint_var(self.last_object_id)

        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'TrackStatus':
        # Parse namespace tuple
        namespace_parts = []
        num_parts = buffer.pull_uint_var()
        for _ in range(num_parts):
            part_len = buffer.pull_uint_var()
            part = buffer.pull_bytes(part_len)
            namespace_parts.append(part)
        namespace = b'/'.join(namespace_parts)

        # Parse track name
        track_name_len = buffer.pull_uint_var()
        track_name = buffer.pull_bytes(track_name_len)

        # Parse status info
        status_code = TrackStatusCode(buffer.pull_uint_var())
        last_group_id = buffer.pull_uint_var()
        last_object_id = buffer.pull_uint_var()

        return cls(
            namespace=namespace,
            track_name=track_name,
            status_code=status_code,
            last_group_id=last_group_id,
            last_object_id=last_object_id
        )


@dataclass
class Subscribe(MOQTMessage):
    """SUBSCRIBE message for requesting track data."""
    subscribe_id: int
    track_alias: int
    namespace: Tuple[bytes, ...]
    track_name: bytes
    priority: int
    direction: int  # Ascending/Descending
    filter_type: int
    start_group: Optional[int] = None
    start_object: Optional[int] = None
    end_group: Optional[int] = None
    parameters: Optional[Dict[int, bytes]] = None

    def __post_init__(self):
        self.type = MOQTMessageType.SUBSCRIBE

    def serialize(self) -> bytes:
        buf = Buffer(capacity=64)
        payload = Buffer(capacity=64)

        payload.push_uint_var(self.subscribe_id)
        payload.push_uint_var(self.track_alias)

        # Add namespace as tuple
        payload.push_uint_var(len(self.namespace))
        for part in self.namespace:
            payload.push_uint_var(len(part))
            payload.push_bytes(part)

        payload.push_uint_var(len(self.track_name))
        payload.push_bytes(self.track_name)
        payload.push_uint8(self.priority)
        payload.push_uint8(self.direction)
        payload.push_uint_var(self.filter_type)

        # Add optional start/end fields based on filter type
        if self.filter_type in (3, 4):  # ABSOLUTE_START or ABSOLUTE_RANGE
            payload.push_uint_var(self.start_group or 0)
            payload.push_uint_var(self.start_object or 0)

        if self.filter_type == 4:  # ABSOLUTE_RANGE
            payload.push_uint_var(self.end_group or 0)

        # Add parameters
        parameters = self.parameters or {}
        payload.push_uint_var(len(parameters))
        for param_id, param_value in parameters.items():
            payload.push_uint_var(param_id)
            payload.push_uint_var(len(param_value))
            payload.push_bytes(param_value)

        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'Subscribe':
        subscribe_id = buffer.pull_uint_var()
        track_alias = buffer.pull_uint_var()

        # Deserialize namespace tuple
        tuple_len = buffer.pull_uint_var()
        namespace = []
        for _ in range(tuple_len):
            part_len = buffer.pull_uint_var()
            namespace.append(buffer.pull_bytes(part_len))
        namespace = tuple(namespace)

        # Track name
        track_name_len = buffer.pull_uint_var()
        track_name = buffer.pull_bytes(track_name_len)

        priority = buffer.pull_uint8()
        direction = buffer.pull_uint8()
        filter_type = buffer.pull_uint_var()

        # Handle optional fields based on filter type
        start_group = None
        start_object = None
        end_group = None
        if filter_type in (3, 4):  # ABSOLUTE_START or ABSOLUTE_RANGE
            start_group = buffer.pull_uint_var()
            start_object = buffer.pull_uint_var()

        if filter_type == 4:  # ABSOLUTE_RANGE
            end_group = buffer.pull_uint_var()

        # Deserialize parameters
        params = {}
        param_count = buffer.pull_uint_var()
        for _ in range(param_count):
            param_id = buffer.pull_uint_var()
            param_len = buffer.pull_uint_var()
            param_value = buffer.pull_bytes(param_len)
            params[param_id] = param_value

        return cls(
            subscribe_id=subscribe_id,
            track_alias=track_alias,
            namespace=namespace,
            track_name=track_name,
            priority=priority,
            direction=direction,
            filter_type=filter_type,
            start_group=start_group,
            start_object=start_object,
            end_group=end_group,
            parameters=params
        )


@dataclass
class Unsubscribe(MOQTMessage):
    """UNSUBSCRIBE message for ending track subscription."""
    subscribe_id: int

    def __post_init__(self):
        self.type = MOQTMessageType.UNSUBSCRIBE

    def serialize(self) -> bytes:
        buf = Buffer(capacity=8)
        payload = Buffer(capacity=8)

        payload.push_uint_var(self.subscribe_id)

        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'Unsubscribe':
        subscribe_id = buffer.pull_uint_var()
        return cls(subscribe_id=subscribe_id)


@dataclass
class SubscribeDone(MOQTMessage):
    """SUBSCRIBE_DONE message indicating subscription completion."""
    subscribe_id: int
    status_code: int  # SubscribeDoneCode
    stream_count: int
    reason: str

    def __post_init__(self):
        self.type = MOQTMessageType.SUBSCRIBE_DONE

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)

        # Write message type and calculate payload size
        payload_size = 0
        payload_size += 1  # subscribe_id varint
        payload_size += 1  # status_code varint
        payload_size += 1  # stream_count varint
        reason_bytes = self.reason.encode()
        payload_size += 1  # reason length varint
        payload_size += len(reason_bytes)  # reason string

        # Write header
        buf.push_uint_var(self.type)
        buf.push_uint_var(payload_size)

        # Write payload
        buf.push_uint_var(self.subscribe_id)
        buf.push_uint_var(self.status_code)
        buf.push_uint_var(self.stream_count)
        buf.push_uint_var(len(reason_bytes))
        buf.push_bytes(reason_bytes)

        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'SubscribeDone':
        subscribe_id = buffer.pull_uint_var()
        status_code = buffer.pull_uint_var()
        stream_count = buffer.pull_uint_var()
        reason_len = buffer.pull_uint_var()
        reason = buffer.pull_bytes(reason_len).decode()

        return cls(
            subscribe_id=subscribe_id,
            status_code=status_code,
            stream_count=stream_count,
            reason=reason
        )


@dataclass
class MaxSubscribeId(MOQTMessage):
    """MAX_SUBSCRIBE_ID message setting maximum subscribe ID."""
    subscribe_id: int

    def __post_init__(self):
        self.type = MOQTMessageType.MAX_SUBSCRIBE_ID

    def serialize(self) -> bytes:
        buf = Buffer(capacity=16)

        payload_size = 1  # subscribe_id varint

        buf.push_uint_var(self.type)
        buf.push_uint_var(payload_size)
        buf.push_uint_var(self.subscribe_id)

        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'MaxSubscribeId':
        subscribe_id = buffer.pull_uint_var()
        return cls(subscribe_id=subscribe_id)


@dataclass
class SubscribesBlocked(MOQTMessage):
    """SUBSCRIBES_BLOCKED message indicating subscriber is blocked."""
    maximum_subscribe_id: int

    def __post_init__(self):
        self.type = MOQTMessageType.SUBSCRIBES_BLOCKED

    def serialize(self) -> bytes:
        buf = Buffer(capacity=16)

        payload_size = 1  # maximum_subscribe_id varint

        buf.push_uint_var(self.type)
        buf.push_uint_var(payload_size)
        buf.push_uint_var(self.maximum_subscribe_id)

        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'SubscribesBlocked':
        maximum_subscribe_id = buffer.pull_uint_var()
        return cls(maximum_subscribe_id=maximum_subscribe_id)


@dataclass
class SubscribeOk(MOQTMessage):
    """SUBSCRIBE_OK message indicating successful subscription."""
    subscribe_id: int
    expires: int
    group_order: int  # 0x1=Ascending, 0x2=Descending
    content_exists: int  # 0 or 1
    largest_group_id: Optional[int] = None  # Only if content_exists=1
    largest_object_id: Optional[int] = None  # Only if content_exists=1
    parameters: Optional[Dict[int, bytes]] = None

    def __post_init__(self):
        self.type = MOQTMessageType.SUBSCRIBE_OK

    def serialize(self) -> bytes:
        buf = Buffer(capacity=64)

        # Calculate payload size
        payload_size = 0
        payload_size += 1  # subscribe_id varint
        payload_size += 1  # expires varint
        payload_size += 1  # group_order uint8
        payload_size += 1  # content_exists uint8

        if self.content_exists == 1:
            payload_size += 1  # largest_group_id varint
            payload_size += 1  # largest_object_id varint

        parameters = self.parameters or {}
        payload_size += 1  # parameter count varint
        for param_id, param_value in parameters.items():
            payload_size += 1  # param id varint
            payload_size += 1  # param length varint
            payload_size += len(param_value)  # param value

        # Write message
        buf.push_uint_var(self.type)
        buf.push_uint_var(payload_size)

        # Write payload fields
        buf.push_uint_var(self.subscribe_id)
        buf.push_uint_var(self.expires)
        buf.push_uint8(self.group_order)
        buf.push_uint8(self.content_exists)

        if self.content_exists == 1:
            buf.push_uint_var(self.largest_group_id or 0)
            buf.push_uint_var(self.largest_object_id or 0)

        # Write parameters
        buf.push_uint_var(len(parameters))
        for param_id, param_value in parameters.items():
            buf.push_uint_var(param_id)
            buf.push_uint_var(len(param_value))
            buf.push_bytes(param_value)

        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'SubscribeOk':
        subscribe_id = buffer.pull_uint_var()
        expires = buffer.pull_uint_var()
        group_order = buffer.pull_uint8()
        content_exists = buffer.pull_uint8()

        largest_group_id = None
        largest_object_id = None
        if content_exists == 1:
            largest_group_id = buffer.pull_uint_var()
            largest_object_id = buffer.pull_uint_var()

        param_count = buffer.pull_uint_var()
        parameters = {}
        for _ in range(param_count):
            param_id = buffer.pull_uint_var()
            param_len = buffer.pull_uint_var()
            param_value = buffer.pull_bytes(param_len)
            parameters[param_id] = param_value

        return cls(
            subscribe_id=subscribe_id,
            expires=expires,
            group_order=group_order,
            content_exists=content_exists,
            largest_group_id=largest_group_id,
            largest_object_id=largest_object_id,
            parameters=parameters
        )


@dataclass
class SubscribeError(MOQTMessage):
    """SUBSCRIBE_ERROR message indicating subscription failure."""
    subscribe_id: int
    error_code: int  # SubscribeErrorCode
    reason: str
    track_alias: int

    def __post_init__(self):
        self.type = MOQTMessageType.SUBSCRIBE_ERROR

    def serialize(self) -> bytes:
        buf = Buffer(capacity=64)

        # Calculate payload size
        reason_bytes = self.reason.encode()
        payload_size = 0
        payload_size += 1  # subscribe_id varint
        payload_size += 1  # error_code varint
        payload_size += 1  # reason length varint
        payload_size += len(reason_bytes)  # reason string
        payload_size += 1  # track_alias varint

        # Write message
        buf.push_uint_var(self.type)
        buf.push_uint_var(payload_size)

        # Write payload fields
        buf.push_uint_var(self.subscribe_id)
        buf.push_uint_var(self.error_code)
        buf.push_uint_var(len(reason_bytes))
        buf.push_bytes(reason_bytes)
        buf.push_uint_var(self.track_alias)

        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'SubscribeError':
        subscribe_id = buffer.pull_uint_var()
        error_code = buffer.pull_uint_var()
        reason_len = buffer.pull_uint_var()
        reason = buffer.pull_bytes(reason_len).decode()
        track_alias = buffer.pull_uint_var()

        return cls(
            subscribe_id=subscribe_id,
            error_code=error_code,
            reason=reason,
            track_alias=track_alias
        )


@dataclass
class SubscribeUpdate(MOQTMessage):
    """SUBSCRIBE_UPDATE message for modifying an existing subscription."""
    subscribe_id: int
    start_group: int
    start_object: int
    end_group: int
    priority: int
    parameters: Optional[Dict[int, bytes]] = None

    def __post_init__(self):
        self.type = MOQTMessageType.SUBSCRIBE_UPDATE

    def serialize(self) -> bytes:
        buf = Buffer(capacity=64)

        # Calculate payload size
        payload_size = 0
        payload_size += 1  # subscribe_id varint
        payload_size += 1  # start_group varint
        payload_size += 1  # start_object varint
        payload_size += 1  # end_group varint
        payload_size += 1  # priority uint8

        parameters = self.parameters or {}
        payload_size += 1  # parameter count varint
        for param_id, param_value in parameters.items():
            payload_size += 1  # param id varint
            payload_size += 1  # param length varint
            payload_size += len(param_value)  # param value

        # Write message
        buf.push_uint_var(self.type)
        buf.push_uint_var(payload_size)

        # Write payload fields
        buf.push_uint_var(self.subscribe_id)
        buf.push_uint_var(self.start_group)
        buf.push_uint_var(self.start_object)
        buf.push_uint_var(self.end_group)
        buf.push_uint8(self.priority)

        # Write parameters
        buf.push_uint_var(len(parameters))
        for param_id, param_value in parameters.items():
            buf.push_uint_var(param_id)
            buf.push_uint_var(len(param_value))
            buf.push_bytes(param_value)

        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'SubscribeUpdate':
        subscribe_id = buffer.pull_uint_var()
        start_group = buffer.pull_uint_var()
        start_object = buffer.pull_uint_var()
        end_group = buffer.pull_uint_var()
        priority = buffer.pull_uint8()

        param_count = buffer.pull_uint_var()
        parameters = {}
        for _ in range(param_count):
            param_id = buffer.pull_uint_var()
            param_len = buffer.pull_uint_var()
            param_value = buffer.pull_bytes(param_len)
            parameters[param_id] = param_value

        return cls(
            subscribe_id=subscribe_id,
            start_group=start_group,
            start_object=start_object,
            end_group=end_group,
            priority=priority,
            parameters=parameters
        )



import inspect
import asyncio
from typing import Optional, Dict, Any, Callable
from collections import defaultdict
import pylsqpack

from aioquic.buffer import Buffer, UINT_VAR_MAX
from aioquic.asyncio.protocol import QuicConnectionProtocol
from aioquic.h3.connection import H3Connection, H3Stream
from aioquic.quic.connection import QuicConnection
from aioquic.quic.events import QuicEvent, StreamDataReceived
from aioquic.h3.events import HeadersReceived, DataReceived
from aioquic.quic.logger import QuicLoggerTrace

from .messages import MOQTMessageHandler
from .types import SessionCloseCode
from .utils.logger import get_logger

logger = get_logger(__name__)


class H3CustomConnection(H3Connection):
    """Custom H3Connection wrapper to support alternate SETTINGS"""

    def __init__(self, quic: QuicConnection, enable_webtransport: bool = False, table_capacity: int = 0) -> None:
        # settings table capacity can be overridden - this should be generalized
        self._max_table_capacity = table_capacity
        self._blocked_streams = 16
        self._enable_webtransport = enable_webtransport

        self._is_client = quic.configuration.is_client
        self._is_done = False
        self._quic = quic
        self._quic_logger: Optional[QuicLoggerTrace] = quic._quic_logger
        self._decoder = pylsqpack.Decoder(
            self._max_table_capacity, self._blocked_streams
        )
        self._decoder_bytes_received = 0
        self._decoder_bytes_sent = 0
        self._encoder = pylsqpack.Encoder()
        self._encoder_bytes_received = 0
        self._encoder_bytes_sent = 0
        self._settings_received = False
        self._stream: Dict[int, H3Stream] = {}

        self._max_push_id: Optional[int] = 8 if self._is_client else None
        self._next_push_id: int = 0

        self._local_control_stream_id: Optional[int] = None
        self._local_decoder_stream_id: Optional[int] = None
        self._local_encoder_stream_id: Optional[int] = None

        self._peer_control_stream_id: Optional[int] = None
        self._peer_decoder_stream_id: Optional[int] = None
        self._peer_encoder_stream_id: Optional[int] = None
        self._received_settings: Optional[Dict[int, int]] = None
        self._sent_settings: Optional[Dict[int, int]] = None

        self._init_connection()
        # report sent settings
        settings = self.sent_settings
        if settings is not None:
            logger.debug("H3 SETTINGS sent:")
            for setting_id, value in settings.items():
                logger.debug(f"  Setting 0x{setting_id:x} = {value}")


class MOQTProtocol(QuicConnectionProtocol):
    """Base MOQT protocol implementation."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._h3: Optional[H3Connection] = None
        self._session_id: Optional[int] = None
        self._control_stream_id: Optional[int] = None
        self._streams: Dict[int, Dict] = {}
        self._groups = defaultdict(lambda: {'objects': 0, 'subgroups': set()})
        self._wt_session = asyncio.Event()
        self._moqt_session = asyncio.Event()
        # Initialize message handling
        self.message_handler = MOQTMessageHandler(self)
        
    def transmit(self) -> None:
        """Transmit pending data."""
        logger.debug("Transmitting data")
        super().transmit()

    def connection_made(self, transport):
        """Called when QUIC connection is established."""
        super().connection_made(transport)
        self._h3 = H3CustomConnection(self._quic, enable_webtransport=True)
        logger.info("H3 connection initialized")

    def quic_event_received(self, event: QuicEvent) -> None:
        """Handle incoming QUIC events."""
        # Log any errors
        if hasattr(event, 'error_code'):
            reason = getattr(event, 'reason_phrase', 'unknown')
            logger.error(
                f"QUIC event: error: {event.error_code} reason: {reason}")
            return

        stream_id = getattr(event, 'stream_id', 'unknown')
        logger.debug(
            f"QUIC event: stream {stream_id}: {event.__class__.__name__}")

        # Debug log any data
        if hasattr(event, 'data'):
            logger.debug(
                f"QUIC event: stream {stream_id}: 0x{event.data.hex()}")

        # Handle MOQT control/data messages
        if isinstance(event, StreamDataReceived):
            if event.stream_id == self._control_stream_id:
                self.message_handler.handle_control_message(event.data)
                return
            elif event.stream_id in self._streams:
                self.message_handler.handle_data_message(event.stream_id, event.data)
                return

        # Pass remaining events to H3
        if self._h3 is not None:
            try:
                for h3_event in self._h3.handle_event(event):
                    self._h3_event_received(h3_event)

                if (isinstance(event, StreamDataReceived) and hasattr(event, 'data') and
                        event.data is not None and event.data[:2] == b'\x00\x04'):
                    logger.debug("H3 SETTINGS received:")
                    # Check received settings
                    settings = self._h3.received_settings
                    if settings is not None:
                        for setting_id, value in settings.items():
                            logger.debug(
                                f"  Setting 0x{setting_id:x} = {value}")

            except Exception as e:
                logger.error(f"QUIC event: error handling event: {e}")
                raise
        else:
            logger.error(
                f"QUIC event: stream {stream_id}: event not handled({event.__class__.__name__})")

    def _h3_event_received(self, event: QuicEvent) -> None:
        """Handle H3-specific events."""
        if isinstance(event, HeadersReceived):
            self._h3_handle_headers_received(event)
        elif isinstance(event, DataReceived):
            self._h3_handle_data_received(event)
        else:
            logger.error(
                f"H3 event: stream {event.stream_id}: {event.__class__.__name__}: event not handled")
            if hasattr(event, 'data') and event.data is not None:
                logger.debug(
                    f"H3 event: stream {event.stream_id}: 0x{event.data.hex()}")

    def _h3_handle_headers_received(self, event: HeadersReceived) -> None:
        """Process incoming H3 headers."""
        status = None
        logger.info(f"H3 event: stream {event.stream_id} HeadersReceived:")

        for name, value in event.headers:
            logger.info(f"  {name.decode()}: {value.decode()}")
            if name == b':status':
                status = value

        stream_id = event.stream_id
        msg = f"H3 event: stream {stream_id}: "
        if status == b"200":
            logger.info(msg + "WebTransport session established")
        else:
            error = f"WebTransport session setup failed ({status})"
            logger.error(msg + error)
            raise Exception(error)
        # signal WebTransport session established event
        self._wt_session.set()

    def _h3_handle_data_received(self, event: DataReceived) -> None:
        """Process incoming H3 data MOQT data is not expected to arrive like this"""
        logger.warning(f"H3 event: stream {event.stream_id}: DataReceived")
        if hasattr(event, 'data'):
            logger.debug(
                f"H3 event: stream {event.stream_id}: data: 0x{event.data.hex()}")

    def register_handler(self, msg_type: int, handler: Callable) -> None:
        """Register a custom message handler."""
        self._custom_handlers[msg_type] = handler

    def send_control_message(self, data: bytes) -> None:
        """Send a MOQT message on the control stream."""
        if self._quic is None or self._control_stream_id is None:
            raise RuntimeError("Control stream not initialized")
        self._quic.send_stream_data(
            stream_id=self._control_stream_id,
            data=data,
            end_stream=False
        )
        self.transmit()

    async def initialize(self, **kwargs) -> None:
        """Initialize the MOQT session. Override in implementation."""
        raise NotImplementedError()

    def close(self, error_code: SessionCloseCode = SessionCloseCode.NO_ERROR,
              reason_phrase: str = "Normal closure") -> None:
        """Close the MOQT session."""
        logger.info(f"Closing MOQT session: {reason_phrase} ({error_code})")

        if self._session_id is not None:
            logger.debug(f"Closing H3 session: stream {self._session_id}")
            self._h3.send_data(self._session_id, b"", end_stream=True)
            self._session_id = None
            self.transmit()

        self._h3 = None

        super().close(error_code, reason_phrase)
        self.transmit()



from enum import IntEnum


class MOQTMessageType:
    """MOQT message type constants."""
    SUBSCRIBE_UPDATE = 0x02
    SUBSCRIBE = 0x03
    SUBSCRIBE_OK = 0x04
    SUBSCRIBE_ERROR = 0x05
    ANNOUNCE = 0x06
    ANNOUNCE_OK = 0x07
    ANNOUNCE_ERROR = 0x08
    UNANNOUNCE = 0x09
    UNSUBSCRIBE = 0x0A
    SUBSCRIBE_DONE = 0x0B
    ANNOUNCE_CANCEL = 0x0C
    TRACK_STATUS_REQUEST = 0x0D
    TRACK_STATUS = 0x0E
    GOAWAY = 0x10
    SUBSCRIBE_ANNOUNCES = 0x11
    SUBSCRIBE_ANNOUNCES_OK = 0x12
    SUBSCRIBE_ANNOUNCES_ERROR = 0x13
    UNSUBSCRIBE_ANNOUNCES = 0x14
    MAX_SUBSCRIBE_ID = 0x15
    FETCH = 0x16
    FETCH_CANCEL = 0x17
    FETCH_OK = 0x18
    FETCH_ERROR = 0x19
    SUBSCRIBES_BLOCKED = 0x1A
    CLIENT_SETUP = 0x40
    SERVER_SETUP = 0x41


class ParamType(IntEnum):
    """Parameter types for MOQT messages."""
    AUTHORIZATION_INFO = 0x02
    DELIVERY_TIMEOUT = 0x03
    MAX_CACHE_DURATION = 0x04


class SetupParamType(IntEnum):
    """Setup Parameter type constants"""
    CLIENT_ROLE = 0x0  # deprecated - removed in draft-8
    ENDPOINT_PATH = 0x1  # only relevant to raw QUIC connection
    MAX_SUBSCRIBER_ID = 0x2


class SessionCloseCode(IntEnum):
    """Session close error codes."""
    NO_ERROR = 0x0
    INTERNAL_ERROR = 0x1
    UNAUTHORIZED = 0x2
    PROTOCOL_VIOLATION = 0x3
    DUPLICATE_TRACK_ALIAS = 0x4
    PARAMETER_LENGTH_MISMATCH = 0x5
    TOO_MANY_SUBSCRIBES = 0x6
    GOAWAY_TIMEOUT = 0x10
    CONTROL_MESSAGE_TIMEOUT = 0x11
    DATA_STREAM_TIMEOUT = 0x12


class SubscribeErrorCode(IntEnum):
    """SUBSCRIBE_ERROR error codes."""
    INTERNAL_ERROR = 0x0
    INVALID_RANGE = 0x1
    RETRY_TRACK_ALIAS = 0x2
    TRACK_DOES_NOT_EXIST = 0x3
    UNAUTHORIZED = 0x4
    TIMEOUT = 0x5


class SubscribeDoneCode(IntEnum):
    """SUBSCRIBE_DONE status codes."""
    UNSUBSCRIBED = 0x0
    INTERNAL_ERROR = 0x1
    UNAUTHORIZED = 0x2
    TRACK_ENDED = 0x3
    SUBSCRIPTION_ENDED = 0x4
    GOING_AWAY = 0x5
    EXPIRED = 0x6
    TOO_FAR_BEHIND = 0x7


class TrackStatusCode(IntEnum):
    """TRACK_STATUS status codes."""
    IN_PROGRESS = 0x0
    DOES_NOT_EXIST = 0x1
    NOT_STARTED = 0x2
    FINISHED = 0x3
    RELAY_NO_INFO = 0x4


class FilterType(IntEnum):
    """Subscription filter types."""
    LATEST_GROUP = 0x1
    RECENT_GROUP = 0x1
    LATEST_OBJECT = 0x2
    ABSOLUTE_START = 0x3
    ABSOLUTE_RANGE = 0x4


class GroupOrder(IntEnum):
    """Group ordering preferences."""
    PUBLISHER_DEFAULT = 0x0
    ASCENDING = 0x1
    DESCENDING = 0x2


class ObjectStatus(IntEnum):
    """Object status codes."""
    NORMAL = 0x0
    DOES_NOT_EXIST = 0x1
    END_OF_GROUP = 0x3
    END_OF_TRACK_AND_GROUP = 0x4
    END_OF_TRACK = 0x5


class StreamType(IntEnum):
    """Stream type identifiers."""
    STREAM_HEADER_SUBGROUP = 0x4
    FETCH_HEADER = 0x5


class ForwardingPreference(IntEnum):
    """Object forwarding preferences."""
    TRACK = 0x0
    SUBGROUP = 0x1
    DATAGRAM = 0x2


class DatagramType(IntEnum):
    """Datagram type identifiers."""
    OBJECT_DATAGRAM = 0x1
    OBJECT_DATAGRAM_STATUS = 0x2



from dataclasses import dataclass
from enum import IntEnum
from typing import Optional, Dict, List, Tuple
from aioquic.buffer import Buffer
from .utils.logger import get_logger
from .types import ObjectStatus, StreamType, ForwardingPreference

logger = get_logger(__name__)


@dataclass
class Track:
    """Represents a MOQT track."""
    namespace: Tuple[bytes, ...]
    name: bytes
    forwarding_preference: ForwardingPreference
    groups: Dict[int, 'Group'] = None

    def __post_init__(self):
        if self.groups is None:
            self.groups = {}

    def add_object(self, obj: 'ObjectHeader') -> None:
        """Add an object to the track's structure."""
        if obj.group_id not in self.groups:
            self.groups[obj.group_id] = Group(group_id=obj.group_id)
        
        self.groups[obj.group_id].add_object(obj)

@dataclass
class Group:
    """Represents a group within a track."""
    group_id: int
    subgroups: Dict[int, 'Subgroup'] = None

    def __post_init__(self):
        if self.subgroups is None:
            self.subgroups = {}

    def add_object(self, obj: 'ObjectHeader') -> None:
        """Add an object to appropriate subgroup."""
        subgroup_id = obj.subgroup_id or 0  # Default to 0 for non-subgroup forwarding
        if subgroup_id not in self.subgroups:
            self.subgroups[subgroup_id] = Subgroup(subgroup_id=subgroup_id)
        
        self.subgroups[subgroup_id].add_object(obj)

@dataclass
class Subgroup:
    """Represents a subgroup within a group."""
    subgroup_id: int
    objects: Dict[int, 'ObjectHeader'] = None

    def __post_init__(self):
        if self.objects is None:
            self.objects = {}

    def add_object(self, obj: 'ObjectHeader') -> None:
        """Add an object to the subgroup."""
        self.objects[obj.object_id] = obj

@dataclass
class ObjectHeader:
    """MOQT object header."""
    track_alias: int
    group_id: int
    object_id: int
    publisher_priority: int
    forwarding_preference: ForwardingPreference
    subgroup_id: Optional[int] = None  # Only used when forwarding_preference is SUBGROUP
    status: ObjectStatus = ObjectStatus.NORMAL
    payload: bytes = b''

    def serialize(self) -> bytes:
        if self.forwarding_preference == ForwardingPreference.DATAGRAM:
            return self.serialize_datagram()
        return self.serialize_stream()

    def serialize_datagram(self) -> bytes:
        """Serialize for datagram transmission."""
        buf = Buffer(capacity=32 + len(self.payload))

        buf.push_uint_var(self.track_alias)
        buf.push_uint_var(self.group_id)
        buf.push_uint_var(self.object_id)
        buf.push_uint8(self.publisher_priority)
        
        if self.payload:
            buf.push_bytes(self.payload)
        elif self.status != ObjectStatus.NORMAL:
            buf.push_uint_var(self.status)

        return buf.data

    def serialize_stream(self) -> bytes:
        """Serialize for stream transmission."""
        buf = Buffer(capacity=32 + len(self.payload))

        if self.forwarding_preference == ForwardingPreference.SUBGROUP:
            buf.push_uint_var(self.object_id)
        else:
            # Track forwarding includes all fields
            buf.push_uint_var(self.track_alias)
            buf.push_uint_var(self.group_id)
            buf.push_uint_var(self.object_id)
            buf.push_uint8(self.publisher_priority)

        # Handle payload/status
        if self.status == ObjectStatus.NORMAL:
            if self.payload:
                buf.push_uint_var(len(self.payload))
                buf.push_bytes(self.payload)
            else:
                buf.push_uint_var(0)
        else:
            buf.push_uint_var(0)  # Zero length
            buf.push_uint_var(self.status)  # Status code

        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer, forwarding_preference: ForwardingPreference, 
                   subgroup_id: Optional[int] = None, stream_header: Optional['StreamHeaderSubgroup'] = None) -> 'ObjectHeader':
        """Deserialize an object based on its forwarding preference."""
        if forwarding_preference == ForwardingPreference.DATAGRAM:
            return cls.deserialize_datagram(buffer)
        elif forwarding_preference == ForwardingPreference.SUBGROUP and stream_header:
            return cls.deserialize_subgroup(buffer, stream_header)
        return cls.deserialize_track(buffer, forwarding_preference, subgroup_id)

    @classmethod
    def deserialize_datagram(cls, buffer: Buffer) -> 'ObjectHeader':
        """Deserialize a datagram object."""
        track_alias = buffer.pull_uint_var()
        group_id = buffer.pull_uint_var()
        object_id = buffer.pull_uint_var()
        publisher_priority = buffer.pull_uint8()
        
        remaining = buffer.pull_bytes(buffer.capacity - buffer.tell())
        if not remaining:
            status = ObjectStatus.NORMAL
            payload = b''
        else:
            try:
                status = ObjectStatus(remaining[0])
                payload = b''
            except ValueError:
                status = ObjectStatus.NORMAL
                payload = remaining

        return cls(
            track_alias=track_alias,
            group_id=group_id,
            object_id=object_id,
            publisher_priority=publisher_priority,
            forwarding_preference=ForwardingPreference.DATAGRAM,
            status=status,
            payload=payload
        )

    @classmethod
    def deserialize_subgroup(cls, buffer: Buffer, header: 'StreamHeaderSubgroup') -> 'ObjectHeader':
        """Deserialize an object within a subgroup stream."""
        object_id = buffer.pull_uint_var()
        payload_len = buffer.pull_uint_var()

        if payload_len == 0:
            try:
                status = ObjectStatus(buffer.pull_uint_var())
                payload = b''
            except ValueError as e:
                logger.error(f"Invalid object status: {e}")
                raise
        else:
            status = ObjectStatus.NORMAL
            payload = buffer.pull_bytes(payload_len)

        return cls(
            track_alias=header.track_alias,
            group_id=header.group_id,
            object_id=object_id,
            publisher_priority=header.publisher_priority,
            forwarding_preference=ForwardingPreference.SUBGROUP,
            subgroup_id=header.subgroup_id,
            status=status,
            payload=payload
        )

    @classmethod
    def deserialize_track(cls, buffer: Buffer, forwarding_preference: ForwardingPreference, 
                         subgroup_id: Optional[int] = None) -> 'ObjectHeader':
        """Deserialize an object with track forwarding."""
        track_alias = buffer.pull_uint_var()
        group_id = buffer.pull_uint_var()
        object_id = buffer.pull_uint_var()
        publisher_priority = buffer.pull_uint8()
        payload_len = buffer.pull_uint_var()

        if payload_len == 0:
            try:
                status = ObjectStatus(buffer.pull_uint_var())
                payload = b''
            except ValueError as e:
                logger.error(f"Invalid object status: {e}")
                raise
        else:
            status = ObjectStatus.NORMAL
            payload = buffer.pull_bytes(payload_len)

        return cls(
            track_alias=track_alias,
            group_id=group_id,
            object_id=object_id,
            publisher_priority=publisher_priority,
            forwarding_preference=forwarding_preference,
            subgroup_id=subgroup_id,
            status=status,
            payload=payload
        )

@dataclass
class StreamHeaderSubgroup:
    """MOQT subgroup stream header."""
    track_alias: int
    group_id: int
    subgroup_id: int
    publisher_priority: int

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)
        buf.push_uint_var(StreamType.STREAM_HEADER_SUBGROUP)
        
        payload = Buffer(capacity=32)
        payload.push_uint_var(self.track_alias)
        payload.push_uint_var(self.group_id)
        payload.push_uint_var(self.subgroup_id)
        payload.push_uint8(self.publisher_priority)

        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'StreamHeaderSubgroup':
        stream_type = buffer.pull_uint_var()
        if stream_type != StreamType.STREAM_HEADER_SUBGROUP:
            raise ValueError(f"Invalid stream type: {stream_type}")

        track_alias = buffer.pull_uint_var()
        group_id = buffer.pull_uint_var()
        subgroup_id = buffer.pull_uint_var()
        publisher_priority = buffer.pull_uint8()

        return cls(
            track_alias=track_alias,
            group_id=group_id,
            subgroup_id=subgroup_id,
            publisher_priority=publisher_priority
        )

@dataclass
class FetchHeader:
    """MOQT fetch stream header."""
    subscribe_id: int

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)
        buf.push_uint_var(StreamType.FETCH_HEADER)
        
        payload = Buffer(capacity=32)
        payload.push_uint_var(self.subscribe_id)

        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'FetchHeader':
        stream_type = buffer.pull_uint_var()
        if stream_type != StreamType.FETCH_HEADER:
            raise ValueError(f"Invalid stream type: {stream_type}")

        subscribe_id = buffer.pull_uint_var()
        return cls(subscribe_id=subscribe_id)

@dataclass
class FetchObject:
    """Object within a fetch stream."""
    group_id: int
    subgroup_id: int 
    object_id: int
    publisher_priority: int
    status: ObjectStatus = ObjectStatus.NORMAL
    payload: bytes = b''

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32 + len(self.payload))
        
        buf.push_uint_var(self.group_id)
        buf.push_uint_var(self.subgroup_id)
        buf.push_uint_var(self.object_id)
        buf.push_uint8(self.publisher_priority)

        if self.status == ObjectStatus.NORMAL:
            buf.push_uint_var(len(self.payload))
            if self.payload:
                buf.push_bytes(self.payload)
        else:
            buf.push_uint_var(0)  # Zero length
            buf.push_uint_var(self.status)  # Status code

        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'FetchObject':
        group_id = buffer.pull_uint_var()
        subgroup_id = buffer.pull_uint_var()
        object_id = buffer.pull_uint_var()
        publisher_priority = buffer.pull_uint8()
        payload_len = buffer.pull_uint_var()

        if payload_len == 0:
            try:
                status = ObjectStatus(buffer.pull_uint_var())
                payload = b''
            except ValueError as e:
                logger.error(f"Invalid object status: {e}")
                raise
        else:
            status = ObjectStatus.NORMAL
            payload = buffer.pull_bytes(payload_len)

        return cls(
            group_id=group_id,
            subgroup_id=subgroup_id,
            object_id=object_id,
            publisher_priority=publisher_priority,
            status=status,
            payload=payload
        )


import asyncio
import ssl
from importlib.metadata import version
from typing import Optional, Union, Tuple, Dict, AsyncContextManager

from aioquic.quic.configuration import QuicConfiguration
from aioquic.asyncio.client import connect
from aioquic.h3.connection import H3_ALPN

from .protocol import MOQTProtocol
from .types import FilterType, GroupOrder, SessionCloseCode, MOQTMessageType, SubscribeErrorCode
from .messages.setup import *
from .messages.announce import *
from .messages.subscribe import *
from .utils.logger import get_logger, QuicDebugLogger

logger = get_logger(__name__)

USER_AGENT = f"aiomoqt-client/{version('aiomoqt')}"


class MOQTClientProtocol(MOQTProtocol):
    """MOQT client implementation."""

    def __init__(self, *args, client: 'MOQTClient', **kwargs):
        super().__init__(*args, **kwargs)
        self._client = client

        # Register handler using closure to capture self
        async def handle_server_setup(msg: MOQTMessage) -> None:
            assert isinstance(msg, ServerSetup)
            logger.info(f"Received ServerSetup: {msg}")

            if self._moqt_session.is_set():
                error = "Received multiple SERVER_SETUP message"
                logger.error(error)
                self.close(
                    error_code=SessionCloseCode.PROTOCOL_VIOLATION,
                    reason_phrase=error
                )
                raise RuntimeError(error)
            # indicate moqt session setup is complete
            self._moqt_session.set()

        # Register the closure as the handler
        self.message_handler.register_handler(
            MOQTMessageType.SERVER_SETUP,
            handle_server_setup
        )

        # Register handler using closure to capture self
        async def handle_subscribe(msg: MOQTMessage) -> None:
            assert isinstance(msg, Subscribe)
            logger.info(f"Received Subscribe: {msg}")

            self.subscribe_ok(msg.subscribe_id)

        # Register the closure as the handler
        self.message_handler.register_handler(
            MOQTMessageType.SUBSCRIBE,
            handle_subscribe
        )

    async def initialize(self) -> None:
        """Initialize WebTransport and MOQT session."""
        # Create WebTransport session
        self._session_id = self._h3._quic.get_next_available_stream_id(
            is_unidirectional=False
        )

        headers = [
            (b":method", b"CONNECT"),
            (b":protocol", b"webtransport"),
            (b":scheme", b"https"),
            (b":authority",
             f"{self._client.host}:{self._client.port}".encode()),
            (b":path", f"/{self._client.endpoint}".encode()),
            (b"sec-webtransport-http3-draft", b"draft02"),
            (b"user-agent", USER_AGENT.encode()),
        ]

        logger.info(
            f"WebTransport connect send: (stream: {self._session_id})")
        self._h3.send_headers(stream_id=self._session_id,
                              headers=headers, end_stream=False)
        self.transmit()
        # Wait for WebTransport session establishment
        try:
            await asyncio.wait_for(self._wt_session.wait(), timeout=30.0)
        except asyncio.TimeoutError:
            logger.error("WebTransport session establishment timeout")
            raise

        # Create MOQT control stream
        self._control_stream_id = self._h3.create_webtransport_stream(
            session_id=self._session_id
        )
        logger.info(f"Created control stream: {self._control_stream_id}")

        # Send CLIENT_SETUP
        logger.info("Sending CLIENT_SETUP")
        self.send_control_message(
            ClientSetup(
                versions=[0xff000007],
                parameters={}
            ).serialize()
        )
        # Wait for SERVER_SETUP
        try:
            await asyncio.wait_for(self._moqt_session.wait(), timeout=10)
            logger.info("MOQT session setup complete")
        except asyncio.TimeoutError:
            logger.error("MOQT session setup timeout")
            raise

    @classmethod
    def _make_namespace_tuple(cls, namespace: Union[str, Tuple[str, ...]]) -> Tuple[bytes, ...]:
        """Convert string or tuple into bytes tuple."""
        if isinstance(namespace, str):
            return tuple(part.encode() for part in namespace.split('/'))
        elif isinstance(namespace, tuple):
            if all(isinstance(x, bytes) for x in namespace):
                return namespace
            return tuple(part.encode() if isinstance(part, str) else part
                         for part in namespace)
        raise ValueError(
            "namespace must be string with '/' delimiters or tuple")

    def subscribe(
        self,
        namespace: str,
        track_name: str,
        subscribe_id: int = 1,
        track_alias: int = 1,
        priority: int = 128,
        group_order: GroupOrder = GroupOrder.ASCENDING,
        filter_type: FilterType = FilterType.LATEST_GROUP,
        start_group: Optional[int] = None,
        start_object: Optional[int] = None,
        end_group: Optional[int] = None,
        parameters: Optional[Dict[int, bytes]] = None
    ) -> None:
        """Subscribe to a track with configurable options."""
        logger.info(f"Subscribing to {namespace}/{track_name}")

        if parameters is None:
            parameters = {}
        namespace_tuple = self._make_namespace_tuple(namespace)
        self.send_control_message(
            Subscribe(
                subscribe_id=subscribe_id,
                track_alias=track_alias,
                namespace=namespace_tuple,
                track_name=track_name.encode(),
                priority=priority,
                direction=group_order,
                filter_type=filter_type,
                start_group=start_group,
                start_object=start_object,
                end_group=end_group,
                parameters=parameters
            ).serialize()
        )

    def subscribe_ok(
        self,
        subscribe_id: int,
        expires: int = 0,  # 0 means no expiry
        group_order: int = GroupOrder.ASCENDING,
        content_exists: int = 0,
        largest_group_id: Optional[int] = None,
        largest_object_id: Optional[int] = None,
        parameters: Optional[Dict[int, bytes]] = None
    ) -> SubscribeOk:
        """Create and send a SUBSCRIBE_OK response."""
        logger.info(f"Sending SUBSCRIBE_OK for subscription {subscribe_id}")

        message = SubscribeOk(
            subscribe_id=subscribe_id,
            expires=expires,
            group_order=group_order,
            content_exists=content_exists,
            largest_group_id=largest_group_id,
            largest_object_id=largest_object_id,
            parameters=parameters or {}
        )
        self.send_control_message(message.serialize())
        return message

    def subscribe_error(
        self,
        subscribe_id: int,
        error_code: int = SubscribeErrorCode.INTERNAL_ERROR,
        reason: str = "Internal error",
        track_alias: Optional[int] = None
    ) -> SubscribeError:
        """Create and send a SUBSCRIBE_ERROR response."""
        logger.info(
            f"Sending SUBSCRIBE_ERROR: sub-id {subscribe_id}: {reason} ({error_code})")

        message = SubscribeError(
            subscribe_id=subscribe_id,
            error_code=error_code,
            reason=reason,
            track_alias=track_alias
        )
        self.send_control_message(message.serialize())
        return message

    def announce(
        self,
        namespace: Union[str, Tuple[str, ...]],
        parameters: Optional[Dict[int, bytes]] = None
    ) -> Announce:
        """Announce track namespace availability."""
        namespace_tuple = self._make_namespace_tuple(namespace)
        logger.info(f"Announcing namespace: {namespace_tuple}")

        message = Announce(
            namespace=namespace_tuple,
            parameters=parameters or {}
        )
        self.send_control_message(message.serialize())
        return message

    def unannounce(
        self,
        namespace: Tuple[bytes, ...]
    ) -> None:
        """Withdraw track namespace announcement."""
        logger.info(f"Unannouncing namespace: {namespace}")

        self.send_control_message(
            Unannounce(
                namespace=namespace
            ).serialize()
        )

    def unsubscribe(
        self,
        subscribe_id: int
    ) -> None:
        """Unsubscribe from a track."""
        logger.info(f"Unsubscribing from subscription {subscribe_id}")

        self.send_control_message(
            Unsubscribe(
                subscribe_id=subscribe_id
            ).serialize()
        )

    def subscribe_announces(
        self,
        namespace_prefix: str,
        parameters: Optional[Dict[int, bytes]] = None
    ) -> None:
        """Subscribe to announcements for a namespace prefix."""
        logger.info(f"Subscribe announces: prefix {namespace_prefix}")

        if parameters is None:
            parameters = {}
        prefix = self._make_namespace_tuple(namespace_prefix)
        self.send_control_message(
            SubscribeAnnounces(
                namespace_prefix=prefix,
                parameters=parameters
            ).serialize()
        )

    def unsubscribe_announces(
        self,
        namespace_prefix: str
    ) -> None:
        """Unsubscribe from announcements for a namespace prefix."""
        logger.info(f"Unsubscribe announces: prefix {namespace_prefix}")
        prefix = self._make_namespace_tuple(namespace_prefix)
        self.send_control_message(
            UnsubscribeAnnounces(
                namespace_prefix=prefix
            ).serialize()
        )


class MOQTClient:  # New connection manager class
    def __init__(
        self,
        host: str,
        port: int,
        endpoint: Optional[str] = None,
        configuration: Optional[QuicConfiguration] = None,
        debug: bool = False
    ):
        self.host = host
        self.port = port
        self.debug = debug
        self.endpoint = endpoint
        if configuration is None:
            configuration = QuicConfiguration(
                alpn_protocols=H3_ALPN,
                is_client=True,
                verify_mode=ssl.CERT_NONE,
                quic_logger=QuicDebugLogger() if debug else None,
                secrets_log_file=open(
                    "/tmp/keylog.client.txt", "a") if debug else None
            )
        self.configuration = configuration
        logger.debug(f"quic_logger: {configuration.quic_logger.__class__}")

    def connect(self) -> AsyncContextManager[MOQTClientProtocol]:
        """Return a context manager that creates MOQTClientProtocol instance."""
        logger.debug(f"MOQTClient: connect: {self.__class__}")
        return connect(
            self.host,
            self.port,
            configuration=self.configuration,
            create_protocol=lambda *args, **kwargs: MOQTClientProtocol(
                *args, **kwargs, client=self)
        )



from .base import MOQTMessage, MOQTMessageHandler

# For convenience, export message types enum
from ..types import MOQTMessageType

__all__ = ['MOQTMessage', 'MOQTMessageHandler', 'MOQTMessageType']



from dataclasses import dataclass
from typing import Dict, Optional, List, Tuple
from aioquic.buffer import Buffer
from .base import MOQTMessage 
from ..types import MOQTMessageType
from ..utils.logger import get_logger

logger = get_logger(__name__)

@dataclass
class Fetch(MOQTMessage):
    """FETCH message to request a range of objects."""
    subscribe_id: int
    namespace: Tuple[bytes, ...]
    track_name: bytes
    subscriber_priority: int
    group_order: int
    start_group: int
    start_object: int
    end_group: int
    end_object: int
    parameters: Dict[int, bytes]

    def __post_init__(self):
        self.type = MOQTMessageType.FETCH

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)
        payload = Buffer(capacity=32)

        payload.push_uint_var(self.subscribe_id)

        # Namespace tuple
        payload.push_uint_var(len(self.namespace))
        for part in self.namespace:
            payload.push_uint_var(len(part))
            payload.push_bytes(part)

        # Track name
        payload.push_uint_var(len(self.track_name))
        payload.push_bytes(self.track_name)

        payload.push_uint8(self.subscriber_priority)
        payload.push_uint8(self.group_order)
        payload.push_uint_var(self.start_group)
        payload.push_uint_var(self.start_object)
        payload.push_uint_var(self.end_group)
        payload.push_uint_var(self.end_object)

        # Parameters
        payload.push_uint_var(len(self.parameters))
        for param_id, param_value in self.parameters.items():
            payload.push_uint_var(param_id)
            payload.push_uint_var(len(param_value))
            payload.push_bytes(param_value)

        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'Fetch':
        subscribe_id = buffer.pull_uint_var()

        # Namespace tuple
        tuple_len = buffer.pull_uint_var()
        namespace = []
        for _ in range(tuple_len):
            part_len = buffer.pull_uint_var()
            namespace.append(buffer.pull_bytes(part_len))

        # Track name
        track_name_len = buffer.pull_uint_var()
        track_name = buffer.pull_bytes(track_name_len)

        subscriber_priority = buffer.pull_uint8()
        group_order = buffer.pull_uint8()
        start_group = buffer.pull_uint_var()
        start_object = buffer.pull_uint_var()
        end_group = buffer.pull_uint_var()
        end_object = buffer.pull_uint_var()

        # Parameters
        params = {}
        param_count = buffer.pull_uint_var()
        for _ in range(param_count):
            param_id = buffer.pull_uint_var()
            param_len = buffer.pull_uint_var()
            param_value = buffer.pull_bytes(param_len)
            params[param_id] = param_value

        return cls(
            subscribe_id=subscribe_id,
            namespace=tuple(namespace),
            track_name=track_name,
            subscriber_priority=subscriber_priority,
            group_order=group_order,
            start_group=start_group,
            start_object=start_object,
            end_group=end_group,
            end_object=end_object,
            parameters=params
        )

@dataclass
class FetchCancel(MOQTMessage):
    """FETCH_CANCEL message to cancel an ongoing fetch."""
    subscribe_id: int

    def __post_init__(self):
        self.type = MOQTMessageType.FETCH_CANCEL

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)
        payload = Buffer(capacity=32)

        payload.push_uint_var(self.subscribe_id)

        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'FetchCancel':
        subscribe_id = buffer.pull_uint_var()
        return cls(subscribe_id=subscribe_id)

@dataclass
class FetchOk(MOQTMessage):
    """FETCH_OK response message."""
    subscribe_id: int
    group_order: int
    end_of_track: int
    largest_group_id: int
    largest_object_id: int
    parameters: Dict[int, bytes]

    def __post_init__(self):
        self.type = MOQTMessageType.FETCH_OK

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)
        payload = Buffer(capacity=32)

        payload.push_uint_var(self.subscribe_id)
        payload.push_uint8(self.group_order)
        payload.push_uint8(self.end_of_track)
        payload.push_uint_var(self.largest_group_id)
        payload.push_uint_var(self.largest_object_id)

        # Parameters
        payload.push_uint_var(len(self.parameters))
        for param_id, param_value in self.parameters.items():
            payload.push_uint_var(param_id)
            payload.push_uint_var(len(param_value))
            payload.push_bytes(param_value)

        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'FetchOk':
        subscribe_id = buffer.pull_uint_var()
        group_order = buffer.pull_uint8()
        end_of_track = buffer.pull_uint8()
        largest_group_id = buffer.pull_uint_var()
        largest_object_id = buffer.pull_uint_var()

        params = {}
        param_count = buffer.pull_uint_var()
        for _ in range(param_count):
            param_id = buffer.pull_uint_var()
            param_len = buffer.pull_uint_var()
            param_value = buffer.pull_bytes(param_len)
            params[param_id] = param_value

        return cls(
            subscribe_id=subscribe_id,
            group_order=group_order,
            end_of_track=end_of_track,
            largest_group_id=largest_group_id,
            largest_object_id=largest_object_id,
            parameters=params
        )

@dataclass
class FetchError(MOQTMessage):
    """FETCH_ERROR response message."""
    subscribe_id: int
    error_code: int
    reason: str

    def __post_init__(self):
        self.type = MOQTMessageType.FETCH_ERROR

    def serialize(self) -> bytes:
        buf = Buffer(capacity=32)
        payload = Buffer(capacity=32)

        payload.push_uint_var(self.subscribe_id)
        payload.push_uint_var(self.error_code)
        
        reason_bytes = self.reason.encode()
        payload.push_uint_var(len(reason_bytes))
        payload.push_bytes(reason_bytes)

        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf.data

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'FetchError':
        subscribe_id = buffer.pull_uint_var()
        error_code = buffer.pull_uint_var()
        reason_len = buffer.pull_uint_var()
        reason = buffer.pull_bytes(reason_len).decode()
        
        return cls(
            subscribe_id=subscribe_id,
            error_code=error_code,
            reason=reason
        )